from abc import ABC
from typing import Dict, ContextManager

from opentelemetry import trace, metrics
from opentelemetry.metrics import UpDownCounter
from opentelemetry.trace.span import Span


class KlymTelemetry(ABC):
    """
    Class wrapping the OpenTelemetry functionality, like metrics and spans. This class serves
    2 functionalities:

    1. It keeps singleton instances of the different metrics created using this class to guarantee
    consistent reporting across the app
    2. It makes sure the same tracer and meter instances are used consistently across the app to
    avoid having instances all over the app

    This is a static class, i.e. all methods are meant to use directly on the class. No instance
    of this class should be created at any point
    """

    __tracer = trace.get_tracer(__name__)
    __meter = metrics.get_meter(__name__)
    __up_down_counters: Dict[str, UpDownCounter] = {}

    @staticmethod
    def __get_or_create_up_down_counter(name: str, description: str) -> UpDownCounter:
        metric = KlymTelemetry.__up_down_counters.get(name, None)
        if not metric:
            metric = KlymTelemetry.__meter.create_up_down_counter(name, description=description)
            KlymTelemetry.__up_down_counters[name] = metric

        return metric

    @staticmethod
    def add_event_curr_span(value: str) -> None:
        """Adds an event to the current span"""
        trace.get_current_span().add_event(value)

    @staticmethod
    def new_curr_span(name: str, **kwargs) -> ContextManager[Span]:   # type: ignore
        """Starts a new span and sets it as the current span"""

        return KlymTelemetry.__tracer.start_as_current_span(name, **kwargs)

    @staticmethod
    def set_attr_curr_span(key: str, value: str) -> None:
        """Sets an attribute to the current span"""
        trace.get_current_span().set_attribute(key, value)

    @staticmethod
    def up(name: str, by: int = 1, description: str = "") -> None:
        """
        Increases the counter with the given name by `by`
        By default, increases the counter by 1
        """

        metric = KlymTelemetry.__get_or_create_up_down_counter(name, description)
        metric.add(by)

    @staticmethod
    def down(name: str, by: int = 1, description: str = "") -> None:
        """
        Reduces the counter with the given name by `-by`
        By default, decreases the counter by 1
        """

        metric = KlymTelemetry.__get_or_create_up_down_counter(name, description)
        metric.add(-1 * by)



from abc import abstractmethod, ABC

from opentelemetry import trace, metrics, propagate
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.propagators.aws import AwsXRayPropagator
from opentelemetry.sdk.extension.aws.trace import AwsXRayIdGenerator
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor


class KLYMInstrumentor(ABC):

    def __init__(self, service_name: str, endpoint: str) -> None:
        propagate.set_global_textmap(AwsXRayPropagator())
        self._service_name = service_name
        self._endpoint = endpoint

    @property
    def _resource(self):
        return Resource.create(attributes={"service.name": self._service_name})

    @property
    def _span_exporter(self):
        return OTLPSpanExporter(endpoint=self._endpoint)

    @property
    def _processor(self):
        return BatchSpanProcessor(span_exporter=self._span_exporter)

    @property
    def _metrics_exporter(self):
        return OTLPMetricExporter(endpoint=self._endpoint)

    @property
    def _reader(self):
        return PeriodicExportingMetricReader(exporter=self._metrics_exporter)

    @property
    def tracer_provider(self):
        tracer_provider = TracerProvider(resource=self._resource, id_generator=AwsXRayIdGenerator())
        tracer_provider.add_span_processor(span_processor=self._processor)
        trace.set_tracer_provider(tracer_provider=tracer_provider)
        return tracer_provider

    @property
    def meter_provider(self):
        meter_provider = MeterProvider(resource=self._resource, metric_readers=[self._reader])
        metrics.set_meter_provider(meter_provider)
        return meter_provider

    @abstractmethod
    def instrument(self, **kwargs):
        pass



from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from opentelemetry.instrumentation.requests import RequestsInstrumentor

from klym_telemetry.instrumenters.base import KLYMInstrumentor


class _FastAPIInstrumentor(KLYMInstrumentor):

    def __init__(self, service_name: str, endpoint: str, **kwargs) -> None:
        super().__init__(service_name=service_name, endpoint=endpoint)
        if "app" not in kwargs:
            raise ValueError("Missing required argument: app")
        self._app = kwargs.pop("app")
        self._excluded_urls = None
        if "excluded_urls" in kwargs:
            self._excluded_urls = kwargs.pop("excluded_urls")

    def instrument(self):
        FastAPIInstrumentor.instrument_app(
            app=self._app,
            tracer_provider=self.tracer_provider,
            meter_provider=self.meter_provider,
            excluded_urls=self._excluded_urls,
        )
        RequestsInstrumentor().instrument()



from klym_telemetry.instrumenters.celery import _CeleryInstrumentor
from klym_telemetry.instrumenters.django import _DjangoInstrumentor
from klym_telemetry.instrumenters.fastapi import _FastAPIInstrumentor
from klym_telemetry.instrumenters.postgres import _Psycopg2Instrumentor

FACTORIES = {
    "fastapi": _FastAPIInstrumentor,
    "django": _DjangoInstrumentor,
    "celery": _CeleryInstrumentor,
    "psycopg2": _Psycopg2Instrumentor,
}


def instrument_app(app_type: str, **kwargs):
    if app_type in FACTORIES:
        return FACTORIES[app_type](**kwargs).instrument()
    raise ValueError(f"Invalid app type: {app_type}")



from opentelemetry.instrumentation.psycopg2 import Psycopg2Instrumentor

from klym_telemetry.instrumenters.base import KLYMInstrumentor


class _Psycopg2Instrumentor(KLYMInstrumentor):

    def __init__(self, service_name: str, endpoint: str) -> None:
        super().__init__(service_name=service_name, endpoint=endpoint)

    def instrument(self):
        Psycopg2Instrumentor().instrument(
            tracer_provider=self.tracer_provider,
            meter_provider=self.meter_provider
        )



from opentelemetry.instrumentation.celery import CeleryInstrumentor
from opentelemetry.instrumentation.requests import RequestsInstrumentor

from klym_telemetry.instrumenters.base import KLYMInstrumentor


class _CeleryInstrumentor(KLYMInstrumentor):

    def __init__(self, service_name: str, endpoint: str) -> None:
        super().__init__(service_name=service_name, endpoint=endpoint)

    def instrument(self):
        CeleryInstrumentor().instrument(
            tracer_provider=self.tracer_provider,
            meter_provider=self.meter_provider
        )
        RequestsInstrumentor().instrument()



from codecs import open
from os import path

from setuptools import setup, find_packages

HERE = path.abspath(path.dirname(__file__))

with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="klym-telemetry",
    version="0.1.3",
    description="Scaffold library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://dummy.readthedocs.io/",
    author="Klym Telemetry",
    author_email="example@email.com",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent"
    ],
    include_package_data=True,
    packages=find_packages('.', exclude=['tests', 'tests.*']),
    install_requires=[
        'opentelemetry-instrumentation-django~=0.38b0',
        'opentelemetry-instrumentation-fastapi~=0.38b0',
        'opentelemetry-instrumentation-requests~=0.38b0',
        'opentelemetry-instrumentation-sqlalchemy~=0.38b0',
        'opentelemetry-instrumentation-psycopg2~=0.38b0',
        'opentelemetry-instrumentation-celery~=0.38b0',
        'opentelemetry-exporter-otlp-proto-grpc~=1.17.0',
        'opentelemetry-propagator-aws-xray~=1.0.1',
        'opentelemetry-sdk-extension-aws~=2.0.1'
    ],
    python_requires='>=3.6',
)
